#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/pci.h>
#include <linux/fs.h>
#include <linux/interrupt.h>
#include <linux/io.h>
#include <linux/device.h>
#include <linux/slab.h>
//#include <linux/sysdev.h>
#include <linux/delay.h>
#include <linux/clk.h>
#include <linux/wait.h>
#include <linux/major.h>
#include <linux/uaccess.h>
#include "dsc_can2_pci.h"

#define LOG(string...) printk(KERN_NOTICE string)

#define DRV_NAME	"DSC-CAN2"
#define DEV_NAME	"DSC-CAN2-PCI"

#define DRV_MAJOR	230
#define RW_BUF_SIZE	256

//#define PCICAN_VENDER_ID        0x1186
//#define PCICAN_DEVICE_ID        0x4300

#define PCICAN_VENDER_ID        0x10ee
#define PCICAN_DEVICE_ID        0x0505

//static unsigned char *read_buf ;
//static unsigned char *write_buf ;

#define MAX_SIZE 100

typedef struct CANMESSAGE
{
  unsigned int  Rcv_msg_count;
  unsigned char Msg_Type;
  unsigned int  Msg_ID;
  unsigned char Msg_len;
  unsigned int  Rtr;
  unsigned char Data[68];
}CANMESSAGE,*PCANMESSAGE;


struct pci_dev *pci_dev;

typedef struct can{
	void *base_addr;
	int device_cnt ;
	struct semaphore sem;
}can_dev;

can_dev candev[2];




typedef struct  {
	unsigned  short offset ;
	unsigned int value ;
} Reg_Val ;

Reg_Val reg_val ;


static int can_init(void * base_addr)
{

	printk("init:baseaddr=%lx",(unsigned long)base_addr);
	writel(0x80, (base_addr+(CFG_STAT * 4))) ;
	writel(0x00, (base_addr+(CFG_STAT  * 4))) ;
	writel(0x1B, (base_addr+(LIMIT * 4))) ;
	writel(0xFF, (base_addr+(RTIE * 4))) ;
	return 0;

}



static int pci_dev_open(struct inode* inode, struct file* file)
{
	int minor ;
	int i=0,j=0;
	CANMESSAGE pCANMsg[100];

	void *base_addr;
	printk("In open\n");
	minor = MINOR(inode->i_rdev);
	if (minor > 2)
	{
		printk(KERN_INFO "%s: Device supports two can\n", DRV_NAME);
		return -1 ;
	}
	file->private_data = &candev[minor];
	base_addr = candev[minor].base_addr ;
	can_init(base_addr);



	// check the minor 0-1
	candev[minor].device_cnt++ ;
	printk(KERN_INFO "%s: Device open successful. Minor=%d\n",DRV_NAME,minor);
	return 0;
}

static int pci_dev_close(struct inode* inode, struct file* file)
{
	
	int minor ;
	void *base_addr ;

	printk("Close the device\n") ;
	minor = MINOR(inode->i_rdev);
	//kfree(candev[minor].pCANMsg);
	candev[minor].device_cnt-- ;
	base_addr = candev[minor].base_addr ;
	//iounmap(base_addr);
	return 0;
}

static ssize_t pci_dev_write(struct file *file, const char __user *user_buf, size_t count, loff_t *f_pos)
{
	int i=0 ,j ;
	unsigned int value ,dlc ,RTR;
	unsigned char write_buf[20];
	unsigned int offset;
	void *base_addr;
	unsigned char MsgType, tmp ;

	can_dev *candev_p = (can_dev*)file->private_data ;
	base_addr = candev_p->base_addr ;

	if(copy_from_user((unsigned char*)(write_buf), user_buf, count))
	{
		return -EFAULT;
	}	

	offset = TCMD * 4;
	value = 0x00;
	writel(value, (base_addr+offset)) ;

	MsgType = write_buf[i] ;
	i++;

	if ( MsgType == 0 ) // Standard CAN Message 
	{
		value = write_buf[i] ;
		i++;
		value = ((write_buf[i]<<8) | value) ;
		i++;

		offset = MSG_ID_1 * 4;
		writel(((value & 0x7F8) >> 3), (base_addr+offset)) ;

		offset = MSG_ID_2 * 4;
		value = ((value & 0x07) << 0x05);
		writel(value, (base_addr+offset)) ;
	}
	else  // Else it is extended Message 
	{
		value = write_buf[i] ;
		i++;
		value = ((write_buf[i]<<8) | value) ;
		i++;
		value = ((write_buf[i]<<16) | value) ;
		i++;
		value = ((write_buf[i]<<24) | value) ;
		i++;

		offset = MSG_ID_1 * 4;
		tmp = (value >> 21 ) ;
		writel(tmp, (base_addr+offset)) ;

		offset = MSG_ID_2 * 4;
		tmp = (value >> 13 ) & 0xff ;
		writel(tmp, (base_addr+offset)) ;

		offset = MSG_ID_3 * 4;
		tmp = (value >> 5 ) & 0xff ;
		writel(tmp, (base_addr+offset)) ;

		offset = MSG_ID_4 * 4;
		tmp = (value & 0x1f) ;
		tmp = (tmp<<3) ;
		writel(tmp, (base_addr+offset)) ;
	}
	

	offset = T_FIR * 4;
	value = write_buf[i];
	dlc = value ;
	i++ ;
	if ( MsgType == 1 ) // Extended CAN Message 
	{
		// For Extended type Messages, set IDE=1
		value = (value | 0x80 );
	}
	// For RTR bit , set RTR = 1 , 0x40 (6th bit to set RTR)
	RTR = write_buf[i];
	i++;
	value = (RTR)? (value | (1<<6)):((value & ~(1<<6)));

	// Set the RTR bit ,writing into register
	writel(value, (base_addr+offset)) ;
	

	offset = 0x4A;
	for ( j=0; j < dlc ; j++ )
	{
		value = write_buf[i] ;
		writel(value, (base_addr+((offset+j)*4))) ;
		i++ ;
	}

//	offset = TCTRL * 4;
//	value = 0x40;
//	writel(value, (base_addr+offset)) ;

	offset = TCMD * 4;
	value = 0x10;
	writel(value, (base_addr+offset)) ;
	*f_pos += count;
	
	return count ;
}


static ssize_t pci_dev_read(struct file* file, char* user_buf, size_t count, loff_t* f_pos)
{
	unsigned int dlc,i, value ;
	unsigned char rx_buf[68];
	unsigned char buf[68];
	unsigned int offset;
	void *base_addr;
        unsigned int can_id=0;
	unsigned char  msgType=0;
	static int rx_count =0;
	CANMESSAGE pCANMsg[10] ;

	can_dev *candev_p = (can_dev*)file->private_data ;
	base_addr = candev_p->base_addr ;

	do{
	for(i=0; i<68; i++)
	{
		rx_buf[i] = 0;
	}
	
	for(i=0; i<5; i++ )
	{
		rx_buf[i] = readl(base_addr + (i * 4));
		
	}
	// Check for IDE Bit 
	if ( (rx_buf[4] & 0x80 ) == 0x80 )
	{
		msgType = 1 ;
		can_id = ((rx_buf[0]<<21));
		can_id = (can_id | (rx_buf[1]<<13));
		can_id = (can_id | (rx_buf[2]<<5));
		can_id = (can_id | (rx_buf[3]>>3));
		pCANMsg[rx_count].Msg_Type = msgType;
		pCANMsg[rx_count].Msg_ID = can_id;
	
	}
	else
	{
		msgType = 0 ;
		can_id = ((rx_buf[0]<<3));
		can_id = (can_id | (rx_buf[1]>>5));
		pCANMsg[rx_count].Msg_Type = msgType;
		pCANMsg[rx_count].Msg_ID = can_id;
	
	}		
	dlc = (rx_buf[4] & (0x0F));
	pCANMsg[rx_count].Rtr = ((rx_buf[4] >> 6) & 0x01);
	pCANMsg[rx_count].Msg_len = dlc;		

	//read data
	offset = RBUF_DATA ;
	for (i=0; i<dlc; i++ )
	{
		buf[i] = readl(base_addr +(( offset +  i) * 4)) ;
		pCANMsg[rx_count].Data[i] = buf[i];
		//printk("buf[i] %x \n",buf[i]);
	
	}
	if(pCANMsg[rx_count].Rtr == 0)
	memcpy(pCANMsg[rx_count].Data,buf,dlc);
	else
        memset(pCANMsg[rx_count].Data,0,dlc);
	
	// Empty the RBUF which is read 
	offset = RCTRL * 4  ;
	value = readl(base_addr + offset ) ;
	
	value = value | 0x10 ;
	writel(value, (base_addr+offset)) ;
	value = readl(base_addr + offset);
	
	value = value & 0x03;  // Masking Receive buffer stats
	count = sizeof(pCANMsg) ;
	
	rx_count++;
	}while(value!=0);

	pCANMsg[0].Rcv_msg_count = rx_count;
	while(rx_count)
	{     
		if(copy_to_user((void*)(user_buf), (&pCANMsg), sizeof(pCANMsg)))
		{
			return -EFAULT;
		}
	rx_count--;	
	}
	*f_pos += count;
	return count; 
}



static int set_baudrate(void *base_addr,int baudrate)
{
	unsigned int Tseg_1;
	unsigned int Tseg_2;
	unsigned int SJW;
	unsigned int S_PRESC;
	unsigned int BitTime_0, BitTime_1, BitTime_2;	

	switch (baudrate)  // setting actual bus timing
	{                           // all @ 16 Mhz
		case 8://CAN_SPEED_10K:
			Tseg_1 = 0x0B;
			Tseg_2 = 2;
			SJW = 1;
			S_PRESC = 63;
			break;
		case 7://CAN_SPEED_20K:
			Tseg_1 = 0x0B;
			Tseg_2 = 0x02;
			SJW = 1;
			S_PRESC = 49;
			break;
		case 6://CAN_SPEED_50K:
			Tseg_1 = 0x0B;
			Tseg_2 = 0x02;
			SJW = 1;
			S_PRESC = 19;
			break;
		case 5://CAN_SPEED_100K:
			Tseg_1 = 0x0B;
			Tseg_2 = 2;
			SJW = 1;
			S_PRESC = 9;
			break;
		case 4://CAN_SPEED_125K:
			Tseg_1 = 0x0B;
			Tseg_2 = 2;
			SJW = 1;
			S_PRESC = 7;
			break;
		case 3://CAN_SPEED_250K:
			Tseg_1 = 0x0B;
			Tseg_2 = 2;
			SJW = 1;
			S_PRESC = 3;
			break;
		case 2://CAN_SPEED_500K:
			Tseg_1 = 0x0B;
			Tseg_2 = 2;
			SJW = 1;
			S_PRESC = 1;
			break;
		case 1://CAN_SPEED_800K:
			Tseg_1 = 0x06;
			Tseg_2 = 1;
			SJW = 1;
			S_PRESC = 1;
			break;
		case 0://CAN_SPEED_1M:
			Tseg_1 = 0x03;
			Tseg_2 = 2;
			SJW = 1;
			S_PRESC = 1;
			break;
		default:
			return -1;
	}

	BitTime_0 = (Tseg_1 ) | (SJW << 6); 
	BitTime_1 = (Tseg_2 ) | ((Tseg_2 - 1) << 5 );
	BitTime_2 = (SJW) | ((Tseg_1) << 4 );


	writel(0x80, (base_addr + (CFG_STAT * 4))) ;
	writel(BitTime_0, (base_addr + (BITTIME0 * 4))) ;
	writel(BitTime_1, (base_addr + (BITTIME1 * 4))) ;
	writel(BitTime_2, (base_addr + (BITTIME2 * 4))) ;
	writel(S_PRESC, (base_addr + (SPRESC * 4))) ;
	writel(S_PRESC, (base_addr + (FPRESC * 4))) ;
	writel(0x00, (base_addr + (CFG_STAT * 4))) ;
	return 0;

}
static long pci_dev_ioctl(struct file* file, unsigned int pci_cmd, unsigned long pci_arg)
{
	int return_value = 0 ;
	unsigned int i =0;	
	char read_buf[20];
	void *base_addr;
	can_dev *candev_p = (can_dev*)file->private_data ;
	base_addr = candev_p->base_addr ;
	

	if(copy_from_user((unsigned char*)(&reg_val), pci_arg, sizeof(reg_val)))
	{
		return -EFAULT;
	}	

	switch ( pci_cmd )
	{
		case 0x11 :
			//read the register and store to reg_val.value 
			reg_val.value = readl(base_addr + (reg_val.offset * 4)) ;
			if(copy_to_user((char*)(pci_arg), (unsigned char*)(&reg_val), sizeof(reg_val)))
			{
				return -EFAULT;
			}	
			break ;

		case 0x12 :
			//read the register and store to reg_val.value
			//Set baudrate	 
			printk(KERN_ERR "IOCTL Base Address : %lx\n", base_addr) ;
			set_baudrate( base_addr, reg_val.value);			
			break;
		
		case 0x15:
			for (i=0; i<15; i++ )
			{
				reg_val.offset = 0x00 ;
				reg_val.value = readl(base_addr+reg_val.offset+(i*4)) ;
				read_buf[i] = reg_val.value ;
			}
			
			if(copy_to_user((char*)(pci_arg), (unsigned char*)(read_buf), i))
			{
				return -EFAULT;
			}		
			break;

		case 0x14 :
			//read the register and store to reg_val.value
			//Read RIF
			reg_val.offset = RTIF;	 
			reg_val.value = readl(base_addr + (reg_val.offset * 4)) ;
			if(copy_to_user((char*)(pci_arg), (unsigned char*)(&reg_val), sizeof(reg_val)))
			{
				return -EFAULT;
			}	
			break ;
		case 0x16 :
			//read the register and store to reg_val.value
			//Read RSTAT
			
			reg_val.offset = RCTRL ;	 
			reg_val.value = readl(base_addr + (reg_val.offset * 4)) ;
			if(copy_to_user((char*)(pci_arg), (unsigned char*)(&reg_val), sizeof(reg_val)))
			{
				return -EFAULT;
			}
			
			break ;

		default:
			break;

	}
	return ( return_value ) ;
}

struct file_operations pci_dev_fops = {
	.owner = THIS_MODULE,
	.open = pci_dev_open,
	.release = pci_dev_close,
	.read = pci_dev_read,
	.write = pci_dev_write,
	.unlocked_ioctl = pci_dev_ioctl

};


static int init_pci(void)
{
	unsigned long mmio_start, mmio_end, mmio_len, mmio_flags;
	void *ioaddr;
	void *base_addr;
	int i;
	pci_dev = pci_get_device(PCICAN_VENDER_ID, PCICAN_DEVICE_ID, NULL);
	if(pci_dev)
	{
		if(pci_enable_device(pci_dev))
		{
			printk(KERN_INFO "Failed to enable device\n");
			return -ENODEV;
		}
		else
		{
			printk( KERN_INFO "Device enabled\n");
		}
	}
	else
	{	
		printk(KERN_INFO "Device not found\n");
		return -ENODEV;
	}
	if(pci_request_regions(pci_dev, "PCI_DRV"))
	{
		LOG("Failed to accquire the memory resources\n");
		goto clean_disable;
	}
	else 
	{
		LOG("Accquired the memory regions\n");
	}

	pci_set_master(pci_dev);

	for(i=0;  i<2; i++)
	{
		mmio_start = pci_resource_start(pci_dev, i);
		mmio_end = pci_resource_end(pci_dev, i);
		mmio_len = pci_resource_len(pci_dev, i);
		mmio_flags = pci_resource_flags(pci_dev, i);
		//mmio_addr =  mmio_start;

		if((IORESOURCE_MEM & mmio_flags))
		{
			LOG("MMIO resource found at region 1\n");
			LOG("Start : %lx \n", mmio_start);
			LOG("End   : %lx \n", mmio_end);
			LOG("Length: %lx \n", mmio_len);
			LOG("Flags : %lx \n", mmio_flags);
		}
		else
		{
			LOG("Region 1 is NOT a MMIO\n");
		}
		
		ioaddr = ioremap(mmio_start, mmio_len);
		if(ioaddr == NULL)
		{
			LOG("Failed to remap the memory\n");
			goto clean_regions;
		}
		else
		{
			LOG("Remapped the memory regions\n");
		}

		base_addr = (long *)ioaddr;
		base_addr = ioaddr;
		candev[i].base_addr = base_addr ;
		candev[i].device_cnt = 0 ;
		printk("baseaddr-%d=%lx",i,(unsigned long)base_addr);
	}
	

	return 0 ;


clean_regions:
	pci_release_regions(pci_dev);
	LOG("Released PCI regions\n");
clean_disable:
	pci_disable_device(pci_dev);
	LOG("Device disabled\n");
	return -1;

} 

static struct class *pci_class;

static int __init pci_driver_init(void)
{
	int status;
	char dev_name[20];
	int i;
	if ( init_pci() < 0 )
	{
		printk("Enable to initialize the PCI device.\n") ;
		return -ENODEV ;
	}

	if((status = register_chrdev(DRV_MAJOR, DRV_NAME, &pci_dev_fops)) < 0)
	{
		printk(KERN_INFO "%s: Failed to register PCI CAN char driver.\n", DRV_NAME);
		return status;
	}

	pci_class = class_create(THIS_MODULE, DRV_NAME);
	if (IS_ERR(pci_class))
		return PTR_ERR(pci_class);

	for(i=0;i<2;i++)
	{
		sprintf(dev_name,"%s-%d",DRV_NAME,i);
		device_create(pci_class, NULL, MKDEV(DRV_MAJOR, i),
                              NULL, dev_name);
	}
	printk(KERN_INFO "%s: Device initialization successful.\n", DRV_NAME);
	return 0;
}

static void __exit pci_driver_cleanup(void)
{
	LOG("Unregistered the Device\n");
	pci_release_regions(pci_dev);
	LOG("Released PCI regions\n");
	pci_disable_device(pci_dev);
	LOG("Device disabled\n");

	/* Remove the device while exiting from the driver*/
	device_destroy(pci_class,  MKDEV(DRV_MAJOR, 0));
	device_destroy(pci_class,  MKDEV(DRV_MAJOR, 1));
	class_destroy(pci_class) ;

	/* Unregister the device */
	unregister_chrdev(DRV_MAJOR, DRV_NAME);
}

module_init(pci_driver_init);
module_exit(pci_driver_cleanup);

MODULE_AUTHOR("Venkat Hallale, Mindteck.");
MODULE_DESCRIPTION("PCI CAN driver");
MODULE_LICENSE("GPL");
